/* Copyright (C) 2002 Jean-Marc Valin */
/**
   @file filters_altivec.h
   @brief Various analysis/synthesis filters (altivec version)
*/
/*
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   
   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
   
   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifdef _USE_ALTIVEC
#include "filters_altivec.h"

/*
// original code
    int i,j;
	float xi,yi;
	for (i=0;i<N;i++)
	{
		xi=x[i];
		y[i] = _num[0]*xi + _mem[0];
		yi=y[i];
		for (j=0;j<ord-1;j++)
		{
			_mem[j] = _mem[j+1] + _num[j+1]*xi - _den[j+1]*yi;
		}
		_mem[ord-1] = _num[ord]*xi - _den[ord]*yi;
	}
*/

void filter_mem2_altivec(const spx_sig_t *x, const spx_coef_t *_num, const spx_coef_t *_den, spx_sig_t *y, int N, int ord, spx_mem_t *_mem)
{
    int i,j;
    float xi, yi;

    __attribute__ ((__aligned__ (16))) float __num[20], __den[20], __mem[20];
    __attribute__ ((__aligned__ (16))) float num[20], den[20], mem[20];
    __attribute__ ((__aligned__ (16))) float y_aligned[4];
    __attribute__ ((__aligned__ (16))) unsigned char permute_mask[16] = 
        { 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3 };

    // Make 16-byte aligned versions of everything.
    memcpy((char *)__num, ((char *)_num) + 4, (ord - 1) * 4);
    memcpy((char *)__den, ((char *)_den) + 4, (ord - 1) * 4);
    memcpy((char *)__mem, ((char *)_mem) + 4, (ord - 1) * 4);

    memcpy((char *)num, ((char *)_num), ord * 4);
    memcpy((char *)den, ((char *)_den), ord * 4);
    memcpy((char *)mem, ((char *)_mem), ord * 4);

    __vector unsigned char zero = vec_splat_u8(0);
    __vector unsigned char vec_permute_mask = vec_ld(0, permute_mask);
    __vector float vec_num0, vec_mem0, vec_mem;
    __vector float vec__mem, vec__num, vec__den;
    __vector float vec_xi, vec_yi, vec_neg_yi;
            
    int pos = 0;
    int startragged = (((unsigned long)x) / 4) & 3;

    // ragged parts cant be sped up.
    if (startragged)
    {
        startragged = 4 - startragged;
        for (i = 0; i < startragged; i++)
        {
            xi=x[i];
            y[i] = _num[0]*xi + _mem[0];
            yi=y[i];
            for (j=0;j<ord-1;j++)
            {
                _mem[j] = _mem[j+1] + _num[j+1]*xi - _den[j+1]*yi;
            }
            _mem[ord-1] = _num[ord]*xi - _den[ord]*yi;                   
        }
        pos = i;
    }
        
    // len gets the total amount of 16-byte aligned stuff I can work on.
    int len = N - startragged;
    int leftovers;
    leftovers = len & 3;
    len -= leftovers;   // 1-3 could be left over for the end.

    vec_num0 = vec_ld(0, num);
    vec_num0 = vec_perm(vec_num0, vec_num0, vec_permute_mask);   // num0 = num[0]

    for (i = pos; i < pos + len; i++)
    {
        vec_mem0 = vec_ld(0, mem);
        vec_mem0 = vec_perm(vec_mem0, vec_mem0, vec_permute_mask);   // num0 = num[0]

        vec_xi = vec_ld(0, &x[i]);
        vec_yi = vec_madd(vec_num0, vec_xi, vec_mem0);
        vec_st(vec_yi, 0, y_aligned);
        memcpy(&y[i], y_aligned, 16);

        xi=x[i];
        yi=y[i];        

        for (j=0;j<ord-1;j++)
        {
            _mem[j] = _mem[j+1] + _num[j+1]*xi - _den[j+1]*yi;
        }
        _mem[ord-1] = _num[ord]*xi - _den[ord]*yi;
    }
    
    pos += len;

    if (leftovers)
    {
        for (i = pos; i < pos + leftovers; i++)
        {
                xi=x[i];
            y[i] = _num[0]*xi + _mem[0];
            yi=y[i];
            for (j=0;j<ord-1;j++)
            {
                _mem[j] = _mem[j+1] + _num[j+1]*xi - _den[j+1]*yi;
            }
            _mem[ord-1] = _num[ord]*xi - _den[ord]*yi;
        }
    }

}
#endif